import json
import numpy as np
from vininfo import Vin
from typing import Any, Dict, List, Optional, Union
from urllib.parse import urljoin
from requests.adapters import HTTPAdapter
from requests import Session
from .config import STANDARD_VAR_NAMES_NHTSA


class ExternalApiSession(Session):
    """
    Custom session for external APIs.

    Inherits from the `requests.Session` class and sets additional headers.
    """
    def __init__(self, *args, **kwargs):
        super(ExternalApiSession, self).__init__(*args, **kwargs)
        self.headers.update({"Accept-Charset": "utf-8"})

class ExternalApiVin:
    """
    Base class for external VIN decoding services.

    Provides methods for looking up VIN information and processing responses.
    """
    def __init__(self): pass
    def lookup_vin(self, vin, **kwargs): 
        """
        Lookup VIN information.

        Args:
            vin (str or list): The VIN(s) to lookup.
            **kwargs: Additional keyword arguments.

        Returns:
            dict or list: The decoded VIN information.

        Raises:
            ValueError: If vin is None.
        """
        pass
    def process_response(self, response): 
        """
        Process the response from the external service.

        Args:
            response: The response from the external service.

        Returns:
            dict: Processed VIN information.
        """
        pass
        
    
class VinNHTSA(ExternalApiVin):
    """
    VIN decoding using the NHTSA (National Highway Traffic Safety Administration) API.

    Provides methods for decoding VINs using the NHTSA API.
    """
    def __init__(self,host="https://vpic.nhtsa.dot.gov/api/vehicles/"):
        super().__init__()
        if host: self.host=host
        self.session = ExternalApiSession()
        self.session.mount(self.url, HTTPAdapter(pool_connections=2, max_retries=5))
            
    @property
    def url(self):
        return self.host
    
    def lookup_vin(self, vin, **kwargs):
        if vin is None: raise ValueError("Vin is required")
        if isinstance(vin, str): response=self.decode_vin(vin, kwargs.get("model_year", None))
        elif isinstance(vin, list): response=self.decode_vin_batch(vin)
        else: return {'wmi':None, 'vds':None, 'vis':None}
        return self.process_response(response)
    
    def standardize(self, data):
        if isinstance(data, dict): 
            return {STANDARD_VAR_NAMES_NHTSA.get(key, key): self.standardize(value) for key, value in data.items()}
        elif isinstance(data, list): return [self.standardize(item) for item in data]
        else: return data
        
    def process_response(self, response):
        
        if (not response) or (response.status_code >= 400): return {}
        
        response = response.json()["Results"]
        if isinstance(response, dict): response=[response]
        details,rdetails=['AxleConfiguration', 'Axles', 'BatteryInfo', 'BodyCabType', 'BrakeSystemType',
                         'CoolingType', 'DriveType', 'EngineConfiguration', 'EngineCylinders', 'EngineManufacturer',
                         'EngineModel', 'FuelInjectionType', 'FuelTypeSecondary', 'Manufacturer', 'OtherEngineInfo', 
                          'Seats', 'Series2', 'TPMS', 'Trim2', 'VehicleType', 'WheelBaseType', 'Wheels'],{}
        results=[]
        for item in response:
            item_stdr = self.standardize(item)
            rdetails={k:item_stdr[k] for k in details if (item_stdr.get(k,None) and ('Not Applicable' not in str(item_stdr.get(k,''))))}
            res={k:v for k,v in {
                "source": 'https://vpic.nhtsa.dot.gov/',
                "maker": item_stdr.get('Make', None),
                "model": item_stdr.get('Model', None),
                "serie": item_stdr.get('Series', None),
                "trim": item_stdr.get('Trim', None),
                "country": item_stdr.get('PlantCountry', None),
                "year": int(item_stdr.get('ModelYear', None)) if item_stdr.get('ModelYear', None) else None,
                "fuel_type": item_stdr.get('FuelTypePrimary', None),
                "transmission_type": item_stdr.get('TransmissionStyle', None),
                "body_type": item_stdr.get('BodyClass', None),
                "other": rdetails if rdetails else None
            }.items() if v}
            if len(res)>2: results.append(res)
        if len(results)==0: return {'wmi':None, 'vds':None, 'vis':None}
        return results[0] if len(results)==1 else results

    def decode_vin(self, vin, model_year=None):
        endpoint = "DecodeVinValues"
        
        if not len(vin) in range(6, 17 + 1): 
            raise ValueError("Vin must be at least 6 characters and at most 17 characters")
        if model_year and model_year < 1981:
            raise ValueError("Model year must be 1981 or later")

        params = {"modelyear": model_year} if model_year else {}
        params["format"] = "json"

        api = urljoin(self.url, f"{endpoint}/{vin}")
        response = self.session.get(api, params=params)

        return response

    def decode_vin_batch(self, vins):
        endpoint = "DecodeVINValuesBatch"
        
        if not len(vins) in range(1, 50 + 1):
            raise ValueError("Pass at least one VIN, and at most 50 VINs")
              
        data={"data": ";".join(vins), "format": "json"}

        api = urljoin(self.url, endpoint)
        response = self.session.post(url=api, data=data)

        return response


class VinInfo(ExternalApiVin):
    """
    VIN decoding using the vininfo Python library.

    Provides methods for decoding VINs using the vininfo library.
    """
    def __init__(self):
        super().__init__()
    
    def lookup_vin(self, vin, **kwargs):
        if vin is None: 
            raise ValueError("Vin is required")
        if isinstance(vin, str): 
            return self.process_response(Vin(vin))
        elif isinstance(vin, list):
            result=[]
            for v in vin:
                result.append(self.process_response(Vin(v)))
            return result
        return {'wmi':None, 'vds':None, 'vis':None}
    
    def process_response(self, response):
        if not response: return {}
        details,rdetails=['body', 'engine', 'plant', 'transmission'],{}
        if response.details: rdetails={k:getattr(response.details, k).name for k in details}
        try: years=int(response.years[0]) if isinstance(response.years, list) else response.years
        except: years=None
        return {"source": "vininfo_python",
                 "wmi": response.wmi,
                 "model": response.details.model.name if rdetails else None,
                 "serie": response.details.serial.name if rdetails else None,
                 "vds": response.vds,
                 "vis": response.vis,
                 "maker": response.manufacturer,
                 "country": response.country,
                 "year": years,
                 "other": {k:v for k,v in {"region": response.region, **rdetails}.items() if v is not None}}



import logging
from .vin_decoder import *

class QVin(VinDecoder):
    """
    VIN decoding and data management class with additional features.

    Provides methods for decoding VINs, looking up VIN information, and exporting data to Excel. Extends the VinDecoder class.
    """
    def __init__(self, credentials=None): 
        if credentials: 
            from databricks import sql
            logging.getLogger("databricks.sql").setLevel(logging.ERROR)
            connector = DatabricksSQLClient(**credentials).execute_statement
        else: 
            from pyspark.sql import SparkSession
            connector = SparkSession.builder.getOrCreate().sql
        super(QVin, self).__init__(connector)

    def lookup_vin(self, vin, query=None): 
        """
        Look up VIN information and replace NaN values with None.

        Args:
            vin (str, list): The VIN(s) to look up.
            query (str): An optional SQL query to filter data from the database.

        Returns:
            pd.DataFrame or None: The decoded VIN information with NaN values replaced by None.
        """
        response=super().lookup_vin(vin, query=query).replace(np.nan, None)
        return response
    
    def get_distinct(self, columns):
        """
        Get distinct values from the database for the specified columns.

        Args:
            columns (str, list): The columns for which to get distinct values.

        Returns:
            dict: A dictionary with columns as keys and arrays of distinct values as values.
        """
        if isinstance(columns, str): columns=[columns]
        return {c:self._distinct(self._db_connector, c) for c in columns}


from databricks import sql

class DatabricksSQLClient:
    """
    Databricks remote SQL client for executing SQL statements on a Databricks cluster.

    This class provides a convenient way to connect to a Databricks cluster and execute SQL statements.

    Args:
        server_hostname (str): The hostname of the Databricks server.
        http_path (str): The HTTP path for the Databricks server.
        access_token (str): The access token for authentication.
    """
    def __init__(self, server_hostname: str, http_path: str, access_token: str):
        self._server_hostname = server_hostname
        self._http_path = http_path
        self._token = access_token

    def execute_statement(self, statement: str):
        """
        Execute an SQL statement on the Databricks cluster.

        Args:
            statement (str): The SQL statement to execute.

        Returns:
            ArrowTable: The result of the SQL query as an Arrow table.
        """
        with sql.connect(server_hostname=self._server_hostname,
                         http_path=self._http_path,
                         access_token=self._token) as connection:

            with connection.cursor() as cursor:
                cursor.execute(statement)
                return cursor.fetchall_arrow()


STANDARD_VAR_NAMES_NHTSA = {
    "ID": "Id",
    "GCWR": "GCWRFrom",
    "GCWR_to": "GCWRTo",
    "GVWR": "GVWRFrom",
    "GVWR_to": "GVWRTo",
    "Make_ID": "MakeId",
    "MakeID": "MakeId",
    "Make_Name": "Make",
    "MakeName": "Make",
    "Mfr_CommonName": "ManufacturerCommonName",
    "Mfr_ID": "ManufacturerId",
    "Mfr_Name": "Manufacturer",
    "ManufacturerName": "Manufacturer",
    "MfrId": "ManufacturerId",
    "MfrName": "Manufacturer",
    "Model_ID": "ModelId",
    "Model_Name": "Model",
    "ModelName": "Model",
    "ModelID": "ModelId",
    "VehicleTypeName": "VehicleType",
}


import setuptools

setuptools.setup(
    name = 'qsvin', 
    packages = ['qsvin'],
    version='0.0.9',
    author='Yaroslav Mavliutov',
    author_email='yaroslavm@questarauto.com',
    description = 'Package provides functionality for decoding vin data',
    url='https://github.com/saferide-tech/QuestarVin',
    license='MIT',
    keywords = ['vin'],
    install_requires=[
        'databricks-sql-connector',
        'requests',
        'vininfo',
        'openpyxl'
      ],
)


import pandas as pd

class QDBConnector:
    """
    Base class for database connectors.

    Provides common methods for working with a database table.
    """

    def selectObj(self, columns, condition):
        """
        Select data from the table.

        Args:
            columns (str or list): The columns to select. Pass "*" to select all columns.
            condition (str): The condition for the selection.

        Returns:
            pd.DataFrame, pyspark DataFrame: The selected data.

        Note:
            - If condition is a list, it will be joined with "&" as the condition.
            - If columns is "*", all columns are selected.
        """
        pass

    def deleteObj(self, idx):
        """
        Delete data from the table based on the provided index.

        Args:
            idx: The index of the data to be deleted.

        Note:
            This method should be implemented in derived classes.
        """
        pass

    def insertObj(self, new_data):
        """
        Insert new data into the table.

        Args:
            new_data (dict): A dictionary containing the data to be inserted.

        Note:
            This method should be implemented in derived classes.
        """
        pass

    def updateObj(self, new_data, index=None):
        """
        Update data in the table.

        Args:
            new_data (dict): A dictionary containing the new data.
            index: The index or condition to specify which data to update.

        Note:
            This method should be implemented in derived classes.
        """
        pass


class QPDConnector(QDBConnector):
    """
    Connector for Pandas DataFrames.

    Provides methods for working with Pandas DataFrames as a database table.
    """

    def __init__(self, table):
        self.table = table

    def selectObj(self, columns, condition):
        if isinstance(condition, list):
            condition = " & ".join(condition)
        return self.table.query(condition)[self.table.columns if columns == "*" else columns].copy()

    def deleteObj(self, idx):
        self.table.drop(idx, inplace=True)

    def insertObj(self, new_data):
        new_data = pd.DataFrame([new_data])
        self.table = pd.concat([self.table, new_data], ignore_index=True)

    def updateObj(self, new_data, index=None):
        if isinstance(new_data, dict):
            new_data = pd.DataFrame(new_data, index=index)
        self.table.update(new_data)

    def columns(self):
        return self.table.columns


class QHiveConnector(QDBConnector):
    """
    Connector for Hive tables.

    Provides methods for working with Hive tables.
    """

    def __init__(self, table, connector, to_pandas=True):
        self._table = table
        self._connector = connector
        self._columns = self._get_columns(self._connector(f"select * from {self._table} limit 1"))

    def _get_columns(self, row):
        if type(row).__name__ == 'Table':
            row = row.to_pandas()
        return row.columns

    def selectObj(self, columns="*", condition=None):
        print(f"select {columns} from {self._table} {'where ' + condition if condition else ''}")
        return self._connector(f"select {columns} from {self._table} {'where ' + condition if condition else ''}")

    def deleteObj(self, condition):
        print(f"delete from {self._table} where {condition}")
        self._connector(f"delete from {self._table} where {condition}")

    def insertObj(self, new_data):
        keys = new_data.keys()
        values = [qstr(x) for x in new_data.values()]
        print(f"insert into {self._table} ({','.join(keys)}) VALUES ({','.join(values)})")
        self._connector(f"insert into table {self._table} ({','.join(keys)}) VALUES ({','.join(values)})")

    def executeQuery(self, query):
        return self._connector(query)

    @property
    def columns(self):
        return self._columns

    @property
    def table(self):
        return self._table



from .db_client import *
from .external_api import *


def insertValsToHive(data):
    """
    Convert data to a list of SQL values for insertion into Hive.

    Args:
        data (dict): Data to be inserted.

    Returns:
        list: List of SQL values.
    """
    vals=[]
    for k,v in data.items():
        if k in ('other','year'): continue
        vals.append(f"\'{v}\'") if v else vals.append(f"NULL")
    return vals

class VinDecoder:
    """
    VIN decoding and data management class.

    Provides methods for decoding VINs, looking up VIN information, and exporting data to Excel.
    """
    def __init__(self, connector):
        self._db_connector =  QHiveConnector("qirevin.out_db", connector)
        self._vin_parser = VinInfo()
        self._vin_external_api=[VinNHTSA()]

    def _spark2pandas(self, df):
        """
        Convert a Spark DataFrame to a Pandas DataFrame.

        Args:
            df: Spark DataFrame.

        Returns:
            pd.DataFrame: Pandas DataFrame.
        """
        return df.to_pandas() if type(df).__name__ == 'Table' else df.toPandas()
    
    def _decomposeVin(self, vin, only_keys=False):
        """
        Decompose a VIN into its components.

        Args:
            vin (str): The VIN to decompose.
            only_keys (bool): If True, return only keys.

        Returns:
            dict: Decomposed VIN information.
        """
        decomposition=self._vin_parser.lookup_vin(vin)
        return {'wmi':decomposition['wmi'], 'vds':decomposition['vds'], 'vis':decomposition['vis']} if only_keys else decomposition
    
    def _queryData(self, vin):
        """
        Query data from the database for the given VINs.

        Args:
            vin (list): List of VINs to query.

        Returns:
            pd.DataFrame: Resulting data.
        """
        lookup_request=''
        for v in vin:
            vin_decoded=self._decomposeVin(v, only_keys=True)
            conditions=f"(vds = '{vin_decoded['vds']}' and wmi = '{vin_decoded['wmi']}' and vis = '{vin_decoded['vis']}')"
            lookup_request = lookup_request + " or " + conditions if lookup_request else conditions
        return self._spark2pandas(self._db_connector.selectObj("*", lookup_request))
        
    def lookup_vin(self, vin, query=None):
        """
        Look up VIN information in the database or through external APIs.

        Args:
            vin (str, list): The VIN(s) to look up.
            query (str): An optional SQL query to filter data from the database.

        Returns:
            pd.DataFrame or None: The decoded VIN information.
        """
        if all([vin,query]) or not any([vin,query]): Exception('ValueError. Use either vin or query')
        if query: return self._spark2pandas(self._db_connector.selectObj("*", query))
        if isinstance(vin, str): vin=[vin]
        result=[self._queryData(vin[i:i+50]) for i in range(0, len(vin), 50)]
        merged_df = pd.concat(result).merge(pd.DataFrame([{**self._decomposeVin(v, only_keys=True), "vin":v} for v in vin]), on=['wmi', 'vds', 'vis'], how='right', indicator=True)
        vin = merged_df[merged_df['_merge'] == 'right_only'].drop(columns=['_merge'])['vin'].values
        for v in vin:
            try:
                decoded_vin = self._parse_vin(v)
                if decoded_vin is not None: result.append(decoded_vin)
            except Exception as e: 
                print(e)
                continue
        if len(result)==0: return None
        elif len(result)==1: return result[0]
        return pd.concat(result)

    def _parse_vin(self, v):
        """
        Parse VIN information and insert it into the database.

        Args:
            v (str): The VIN to parse.

        Returns:
            pd.DataFrame or None: The parsed VIN information.
        """
        vin_info=self._decomposeVin(v)
        conditions=f"vds = '{vin_info['vds']}' and wmi = '{vin_info['wmi']}' and vis = '{vin_info['vis']}'"
        external_api_res = {}
        for api in self._vin_external_api:
            try: external_api_res.update(api.lookup_vin(v))
            except: continue
        insert_data={**vin_info, **external_api_res, "other": {**vin_info.get("other", {}), **external_api_res.get("other", {})}} if external_api_res else vin_info
        self._insert({k:v for k,v in insert_data.items() if v is not None}, self._db_connector)
        return self._spark2pandas(self._db_connector.selectObj("*", conditions))
    
    def export_excel(self, vin, query=None, hierarchical=False, fname="output.xlsx"):
        """
        Export VIN information to an Excel file.

        Args:
            vin (str, list): The VIN(s) to export.
            query (str): An optional SQL query to filter data from the database.
            fname (str): The output file name.
        """
        response = self.lookup_vin(vin=vin, query=query, hierarchical=hierarchical)
        if isinstance(response, pd.DataFrame): 
            response['date_created'] = response['date_created'].dt.tz_localize(None)
            response.to_excel(fname)
        else: Exception('ValueError. No data to write')

    def _distinct(self, table, columns):
        """
        Get distinct values from a database table.

        Args:
            table: The database table to query.
            columns (str, list): The columns for which to get distinct values.

        Returns:
            np.ndarray: Array of distinct values.
        """
        if isinstance(columns, str): columns=[columns]
        q = f"select distinct {','.join(columns)} from {table.table}"
        return self._spark2pandas(self._db_connector.executeQuery(q)).values[:, 0]

    def _insert(self, data,table):
        data={**{k:None for k in table.columns if k != "date_created"}, **data}
        vals=','.join(insertValsToHive(data))
        q=f"""insert into table {table.table} ({','.join([x for x in data.keys() if x not in ('other', 'year', 'date_created')])},other,year,date_created) VALUES ({vals}, '{json.dumps(data.get('other', {}))}', {data.get("year", None)},current_timestamp())"""
        print(q)
        self._db_connector.executeQuery(q)